@@ -26,6 +26,7 @@
 
 #include <glib.h>
 #include <glib/gi18n.h>
+#include <gdk/gdkkeysyms.h>
 #include <gtk/gtk.h>
 
 #include <gconf/gconf-client.h>
@@ -62,6 +63,8 @@ struct GvcMixerDialogPrivate
         GtkWidget       *output_treeview;
         GtkWidget       *output_settings_box;
         GtkWidget       *output_balance_bar;
+        GtkWidget       *output_fade_bar;
+        GtkWidget       *output_lfe_bar;
         GtkWidget       *input_treeview;
         GtkWidget       *sound_theme_chooser;
         GtkWidget       *click_feedback_button;
@@ -78,6 +81,7 @@ enum {
         DEVICE_COLUMN,
         ACTIVE_COLUMN,
         ID_COLUMN,
+        SPEAKERS_COLUMN,
         NUM_COLUMNS
 };
 
@@ -95,6 +99,9 @@ static void     bar_set_stream              (GvcMixerDialog      *dialog,
                                              GtkWidget           *bar,
                                              GvcMixerStream      *stream);
 
+static void     on_adjustment_value_changed (GtkAdjustment  *adjustment,
+                                             GvcMixerDialog *dialog);
+
 G_DEFINE_TYPE (GvcMixerDialog, gvc_mixer_dialog, GTK_TYPE_DIALOG)
 
 static void
@@ -139,6 +146,7 @@ update_output_settings (GvcMixerDialog *dialog)
 {
         GvcMixerStream *stream;
         GvcChannelMap  *map;
+        guint           num_settings;
 
         g_debug ("Updating output settings");
         if (dialog->priv->output_balance_bar != NULL) {
@@ -146,6 +154,16 @@ update_output_settings (GvcMixerDialog *dialog)
                                       dialog->priv->output_balance_bar);
                 dialog->priv->output_balance_bar = NULL;
         }
+        if (dialog->priv->output_fade_bar != NULL) {
+                gtk_container_remove (GTK_CONTAINER (dialog->priv->output_settings_box),
+                                      dialog->priv->output_fade_bar);
+                dialog->priv->output_fade_bar = NULL;
+        }
+        if (dialog->priv->output_lfe_bar != NULL) {
+                gtk_container_remove (GTK_CONTAINER (dialog->priv->output_settings_box),
+                                      dialog->priv->output_lfe_bar);
+                dialog->priv->output_lfe_bar = NULL;
+        }
 
         stream = gvc_mixer_control_get_default_sink (dialog->priv->mixer_control);
         if (stream == NULL) {
@@ -159,7 +177,7 @@ update_output_settings (GvcMixerDialog *dialog)
                 return;
         }
 
-        dialog->priv->output_balance_bar = gvc_balance_bar_new (map);
+        dialog->priv->output_balance_bar = gvc_balance_bar_new (map, BALANCE_TYPE_RL);
         if (dialog->priv->size_group != NULL) {
                 gvc_balance_bar_set_size_group (GVC_BALANCE_BAR (dialog->priv->output_balance_bar),
                                                 dialog->priv->size_group,
@@ -169,6 +187,38 @@ update_output_settings (GvcMixerDialog *dialog)
                             dialog->priv->output_balance_bar,
                             FALSE, FALSE, 12);
         gtk_widget_show (dialog->priv->output_balance_bar);
+        num_settings = 1;
+
+        if (gvc_channel_map_can_fade (map)) {
+                dialog->priv->output_fade_bar = gvc_balance_bar_new (map, BALANCE_TYPE_FR);
+                if (dialog->priv->size_group != NULL) {
+                        gvc_balance_bar_set_size_group (GVC_BALANCE_BAR (dialog->priv->output_fade_bar),
+                                                        dialog->priv->size_group,
+                                                        TRUE);
+                }
+                gtk_box_pack_start (GTK_BOX (dialog->priv->output_settings_box),
+                                    dialog->priv->output_fade_bar,
+                                    FALSE, FALSE, 12);
+                gtk_widget_show (dialog->priv->output_fade_bar);
+                num_settings++;
+        }
+
+        if (gvc_channel_map_has_lfe (map)) {
+                dialog->priv->output_lfe_bar = gvc_balance_bar_new (map, BALANCE_TYPE_LFE);
+                if (dialog->priv->size_group != NULL) {
+                        gvc_balance_bar_set_size_group (GVC_BALANCE_BAR (dialog->priv->output_lfe_bar),
+                                                        dialog->priv->size_group,
+                                                        TRUE);
+                }
+                gtk_box_pack_start (GTK_BOX (dialog->priv->output_settings_box),
+                                    dialog->priv->output_lfe_bar,
+                                    FALSE, FALSE, 12);
+                gtk_widget_show (dialog->priv->output_lfe_bar);
+                num_settings++;
+        }
+
+        /* FIXME: We could make this into a "No settings" label instead */
+        gtk_widget_set_sensitive (dialog->priv->output_balance_bar, gvc_channel_map_can_balance (map));
 }
 
 static void
@@ -323,6 +373,7 @@ create_monitor_stream_for_source (GvcMixerDialog *dialog,
         pa_sample_spec ss;
         pa_context    *context;
         int            res;
+        pa_proplist   *proplist;
 
         if (stream == NULL) {
                 return;
@@ -347,7 +398,10 @@ create_monitor_stream_for_source (GvcMixerDialog *dialog,
 
         snprintf (t, sizeof (t), "%u", gvc_mixer_stream_get_index (stream));
 
-        s = pa_stream_new (context, _("Peak detect"), &ss, NULL);
+        proplist = pa_proplist_new ();
+        pa_proplist_sets (proplist, PA_PROP_APPLICATION_ID, "org.gnome.VolumeControl");
+        s = pa_stream_new_with_proplist (context, _("Peak detect"), &ss, NULL, proplist);
+        pa_proplist_free (proplist);
         if (s == NULL) {
                 g_warning ("Failed to create monitoring stream");
                 return;
@@ -374,13 +428,23 @@ on_mixer_control_default_source_changed (GvcMixerControl *control,
                                          GvcMixerDialog  *dialog)
 {
         GvcMixerStream *stream;
+        GtkAdjustment *adj;
 
         g_debug ("GvcMixerDialog: default source changed: %u", id);
 
         stream = gvc_mixer_control_lookup_stream_id (dialog->priv->mixer_control,
                                                      id);
+        adj = GTK_ADJUSTMENT (gvc_channel_bar_get_adjustment (GVC_CHANNEL_BAR (dialog->priv->input_bar)));
+        g_signal_handlers_disconnect_by_func(adj, on_adjustment_value_changed, dialog);
         bar_set_stream (dialog, dialog->priv->input_bar, stream);
-
+        gvc_channel_bar_set_base_volume (GVC_CHANNEL_BAR (dialog->priv->input_bar),
+                                         gvc_mixer_stream_get_base_volume (stream));
+        gvc_channel_bar_set_is_amplified (GVC_CHANNEL_BAR (dialog->priv->input_bar),
+                                          gvc_mixer_stream_get_can_decibel (stream));
+        g_signal_connect (adj,
+                          "value-changed",
+                          G_CALLBACK (on_adjustment_value_changed),
+                          dialog);
         create_monitor_stream_for_source (dialog, stream);
 
         update_default_input (dialog);
@@ -473,7 +537,7 @@ on_adjustment_value_changed (GtkAdjustment  *adjustment,
         stream = g_object_get_data (G_OBJECT (adjustment), "gvc-mixer-dialog-stream"),
         volume = gtk_adjustment_get_value (adjustment);
         if (stream != NULL) {
-                gvc_mixer_stream_change_volume (stream, (guint)volume);
+                gvc_mixer_stream_set_volume(stream, (pa_volume_t) volume );
         }
 }
 
@@ -520,7 +584,8 @@ on_stream_volume_notify (GObject        *object,
         bar = lookup_bar_for_stream (dialog, stream);
 
         if (bar == NULL) {
-                g_warning ("Unable to find bar for stream");
+                g_warning ("Unable to find bar for stream %s in on_stream_volume_notify()",
+                           gvc_mixer_stream_get_name (stream));
                 return;
         }
 
@@ -551,7 +616,8 @@ on_stream_is_muted_notify (GObject        *object,
         bar = lookup_bar_for_stream (dialog, stream);
 
         if (bar == NULL) {
-                g_warning ("Unable to find bar for stream");
+                g_warning ("Unable to find bar for stream %s in on_stream_is_muted_notify()",
+                           gvc_mixer_stream_get_name (stream));
                 return;
         }
 
@@ -639,11 +705,13 @@ add_stream (GvcMixerDialog *dialog,
         gboolean       is_muted;
         gboolean       is_default;
         GtkAdjustment *adj;
+        const char    *id;
 
         g_assert (stream != NULL);
 
         bar = NULL;
         is_default = FALSE;
+        id = gvc_mixer_stream_get_application_id (stream);
 
         if (stream == gvc_mixer_control_get_default_sink (dialog->priv->mixer_control)) {
                 bar = dialog->priv->output_bar;
@@ -657,14 +725,25 @@ add_stream (GvcMixerDialog *dialog,
                 bar = dialog->priv->input_bar;
                 is_default = TRUE;
 
+                /* Disconnect the adj, otherwise it might change if is_amplified changes */
+                adj = GTK_ADJUSTMENT (gvc_channel_bar_get_adjustment (GVC_CHANNEL_BAR (bar)));
+                g_signal_handlers_disconnect_by_func(adj, on_adjustment_value_changed, dialog);
+
+                gvc_channel_bar_set_base_volume (GVC_CHANNEL_BAR (dialog->priv->input_bar),
+                                                 gvc_mixer_stream_get_base_volume (stream));
+                gvc_channel_bar_set_is_amplified (GVC_CHANNEL_BAR (bar),
+                                                  gvc_mixer_stream_get_can_decibel (stream));
+
                 create_monitor_stream_for_source (dialog, stream);
         } else if (stream == gvc_mixer_control_get_event_sink_input (dialog->priv->mixer_control)) {
                 bar = dialog->priv->effects_bar;
                 g_debug ("Adding effects stream");
         } else if (! GVC_IS_MIXER_SOURCE (stream)
                    && !GVC_IS_MIXER_SINK (stream)
-                   && !GVC_IS_MIXER_SOURCE_OUTPUT (stream)
-                   && !gvc_mixer_stream_is_event_stream (stream)) {
+                   && !gvc_mixer_stream_is_event_stream (stream)
+                   && !gvc_mixer_stream_is_virtual (stream)
+                   && g_strcmp0 (id, "org.gnome.VolumeControl") != 0
+                   && g_strcmp0 (id, "org.PulseAudio.pavucontrol") != 0) {
                 bar = create_bar (dialog, dialog->priv->apps_size_group, FALSE);
                 gvc_channel_bar_set_name (GVC_CHANNEL_BAR (bar),
                                           gvc_mixer_stream_get_name (stream));
@@ -690,17 +769,20 @@ add_stream (GvcMixerDialog *dialog,
                                     ID_COLUMN, gvc_mixer_stream_get_id (stream),
                                     -1);
         } else if (GVC_IS_MIXER_SINK (stream)) {
-                GtkTreeModel *model;
-                GtkTreeIter   iter;
+                GtkTreeModel  *model;
+                GtkTreeIter    iter;
+                GvcChannelMap *map;
 
                 model = gtk_tree_view_get_model (GTK_TREE_VIEW (dialog->priv->output_treeview));
                 gtk_list_store_append (GTK_LIST_STORE (model), &iter);
+                map = gvc_mixer_stream_get_channel_map (stream);
                 gtk_list_store_set (GTK_LIST_STORE (model),
                                     &iter,
                                     NAME_COLUMN, gvc_mixer_stream_get_description (stream),
                                     DEVICE_COLUMN, "",
                                     ACTIVE_COLUMN, is_default,
                                     ID_COLUMN, gvc_mixer_stream_get_id (stream),
+                                    SPEAKERS_COLUMN, gvc_channel_map_get_mapping (map),
                                     -1);
         }
 
@@ -913,6 +995,36 @@ on_output_radio_toggled (GtkCellRendererToggle *renderer,
         }
 }
 
+static void
+name_to_text (GtkTreeViewColumn *column,
+              GtkCellRenderer *cell,
+              GtkTreeModel *model,
+              GtkTreeIter *iter,
+              gpointer user_data)
+{
+        char *name, *mapping;
+
+        gtk_tree_model_get(model, iter,
+                           NAME_COLUMN, &name,
+                           SPEAKERS_COLUMN, &mapping,
+                           -1);
+
+        if (mapping == NULL) {
+                g_object_set (cell, "text", name, NULL);
+        } else {
+                char *str;
+
+                str = g_strdup_printf ("%s\n\t<i>%s</i>",
+                                       name, mapping);
+                g_object_set (cell, "markup", str, NULL);
+                g_free (str);
+        }
+
+        g_free (name);
+        g_free (mapping);
+}
+
+
 static GtkWidget *
 create_stream_treeview (GvcMixerDialog *dialog,
                         GCallback       on_toggled)
@@ -929,7 +1041,8 @@ create_stream_treeview (GvcMixerDialog *dialog,
                                     G_TYPE_STRING,
                                     G_TYPE_STRING,
                                     G_TYPE_BOOLEAN,
-                                    G_TYPE_UINT);
+                                    G_TYPE_UINT,
+                                    G_TYPE_STRING);
         gtk_tree_view_set_model (GTK_TREE_VIEW (treeview),
                                  GTK_TREE_MODEL (store));
 
@@ -946,23 +1059,47 @@ create_stream_treeview (GvcMixerDialog *dialog,
                           G_CALLBACK (on_toggled),
                           dialog);
 
-        renderer = gtk_cell_renderer_text_new ();
-        column = gtk_tree_view_column_new_with_attributes (_("Name"),
-                                                           renderer,
-                                                           "text", NAME_COLUMN,
-                                                           NULL);
-        gtk_tree_view_append_column (GTK_TREE_VIEW (treeview), column);
+        gtk_tree_view_insert_column_with_data_func (GTK_TREE_VIEW (treeview), -1,
+                                                    _("Name"), gtk_cell_renderer_text_new (),
+                                                    name_to_text, NULL, NULL);
 
+#if 0
         renderer = gtk_cell_renderer_text_new ();
         column = gtk_tree_view_column_new_with_attributes (_("Device"),
                                                            renderer,
                                                            "text", DEVICE_COLUMN,
                                                            NULL);
         gtk_tree_view_append_column (GTK_TREE_VIEW (treeview), column);
-
+#endif
         return treeview;
 }
 
+static const guint tab_accel_keys[] = {
+        GDK_1, GDK_2, GDK_3, GDK_4
+};
+
+static void
+dialog_accel_cb (GtkAccelGroup    *accelgroup,
+                 GObject          *object,
+                 guint             key,
+                 GdkModifierType   mod,
+                 GvcMixerDialog   *self)
+{
+        gint num = -1;
+        gint i;
+
+        for (i = 0; i < G_N_ELEMENTS (tab_accel_keys); i++) {
+                if (tab_accel_keys[i] == key) {
+                        num = i;
+                        break;
+                }
+        }
+
+        if (num != -1) {
+                gtk_notebook_set_current_page (GTK_NOTEBOOK (self->priv->notebook), num);
+        }
+}
+
 static GObject *
 gvc_mixer_dialog_constructor (GType                  type,
                               guint                  n_construct_properties,
@@ -980,17 +1117,16 @@ gvc_mixer_dialog_constructor (GType                  type,
         GSList           *l;
         GvcMixerStream   *stream;
         GtkTreeSelection *selection;
+        GtkAccelGroup    *accel_group;
+        GClosure         *closure;
+        gint             i;
 
         object = G_OBJECT_CLASS (gvc_mixer_dialog_parent_class)->constructor (type, n_construct_properties, construct_params);
 
         self = GVC_MIXER_DIALOG (object);
         gtk_dialog_add_button (GTK_DIALOG (self), "gtk-close", GTK_RESPONSE_OK);
 
-#if GTK_CHECK_VERSION(2,14,0)
         main_vbox = gtk_dialog_get_content_area (GTK_DIALOG (self));
-#else
-        main_vbox = GTK_DIALOG (self)->vbox;
-#endif
 
         gtk_container_set_border_width (GTK_CONTAINER (self), 6);
 
@@ -1000,7 +1136,7 @@ gvc_mixer_dialog_constructor (GType                  type,
                             FALSE, FALSE, 12);
         self->priv->output_bar = create_bar (self, self->priv->size_group, TRUE);
         gvc_channel_bar_set_name (GVC_CHANNEL_BAR (self->priv->output_bar),
-                                  _("Output volume: "));
+                                  _("Output Volume: "));
         gtk_widget_set_sensitive (self->priv->output_bar, FALSE);
         gtk_box_pack_start (GTK_BOX (self->priv->output_stream_box),
                             self->priv->output_bar, TRUE, TRUE, 12);
@@ -1010,6 +1146,23 @@ gvc_mixer_dialog_constructor (GType                  type,
                             self->priv->notebook,
                             TRUE, TRUE, 6);
 
+        /* Set up accels (borrowed from Empathy) */
+        accel_group = gtk_accel_group_new ();
+        gtk_window_add_accel_group (GTK_WINDOW (self), accel_group);
+
+        for (i = 0; i < G_N_ELEMENTS (tab_accel_keys); i++) {
+                closure =  g_cclosure_new (G_CALLBACK (dialog_accel_cb),
+                                           self,
+                                           NULL);
+                gtk_accel_group_connect (accel_group,
+                                         tab_accel_keys[i],
+                                         GDK_MOD1_MASK,
+                                         0,
+                                         closure);
+        }
+
+        g_object_unref (accel_group);
+
         /* Effects page */
         self->priv->sound_effects_box = gtk_vbox_new (FALSE, 6);
         gtk_container_set_border_width (GTK_CONTAINER (self->priv->sound_effects_box), 12);
@@ -1040,7 +1193,7 @@ gvc_mixer_dialog_constructor (GType                  type,
 
         self->priv->input_bar = create_bar (self, self->priv->size_group, TRUE);
         gvc_channel_bar_set_name (GVC_CHANNEL_BAR (self->priv->input_bar),
-                                  _("Input volume: "));
+                                  _("Input Volume: "));
         gvc_channel_bar_set_low_icon_name (GVC_CHANNEL_BAR (self->priv->input_bar),
                                            "audio-input-microphone-low");
         gvc_channel_bar_set_high_icon_name (GVC_CHANNEL_BAR (self->priv->input_bar),
@@ -1153,7 +1306,7 @@ gvc_mixer_dialog_constructor (GType                  type,
         gtk_notebook_append_page (GTK_NOTEBOOK (self->priv->notebook),
                                   self->priv->applications_box,
                                   label);
-        self->priv->no_apps_label = gtk_label_new (_("No application is currently playing audio."));
+        self->priv->no_apps_label = gtk_label_new (_("No application is currently playing or recording audio."));
         gtk_box_pack_start (GTK_BOX (self->priv->applications_box),
                             self->priv->no_apps_label,
                             TRUE, TRUE, 0);
